import unittest

from webpush_channels import __version__ as VERSION

from .support import BaseWebTest


class HelloViewTest(BaseWebTest, unittest.TestCase):
    def test_returns_info_about_url_and_version(self):
        response = self.app.get('/')
        self.assertEqual(response.json['project_name'], 'webpush-channels')
        self.assertEqual(response.json['project_version'], VERSION)
        self.assertEqual(response.json['project_docs'],
                         'https://webpush-channels-broadcasting.readthedocs.io/')
        self.assertEqual(response.json['url'], 'http://localhost/v0/')


class LoadBalancerHeartbeat(BaseWebTest, unittest.TestCase):
    def test_checks_if_lheartbeat_is_working(self):
        resp = self.app.get('/__lbheartbeat__')
        self.assertEqual(resp.json, {})


class Heartbeat(BaseWebTest, unittest.TestCase):
    def test_returns_storage_true_if_ok(self):
        response = self.app.get('/__heartbeat__')
        self.assertEqual(response.json['storage'], True)

    def test_returns_cache_true_if_ok(self):
        response = self.app.get('/__heartbeat__')
        self.assertEqual(response.json['cache'], True)

    def test_successful_if_one_heartbeat_is_none(self):
        self.app.app.registry.heartbeats['probe'] = lambda r: None
        response = self.app.get('/__heartbeat__', status=200)
        self.assertEqual(response.json['probe'], None)



from webpush_channels.views.subscriptions import SubscriptionSchema
import unittest


class SubscriptionSchemaTest(unittest.TestCase):
    def setUp(self):
        self.schema = SubscriptionSchema()
        self.schema = self.schema.bind()
        keys = dict(auth="authToken",
                    p256dh="encryptionKey")
        self.record = dict(endpoint="https://push.mozilla.com",
                           keys=keys)
        self.deserialized = self.schema.deserialize(self.record)

    def test_record_validation(self):
        self.assertEqual(self.deserialized, self.record)



import unittest
from copy import deepcopy
import json

import mock

from kinto.core import testing
from kinto.core.errors import ERRORS
from kinto.core.storage import exceptions as storage_exceptions

from ..utils import canonical_json
from .support import BaseWebTest, MINIMALIST_SUBSCRIPTION, MINIMALIST_PAYLOAD


class ChannelRegistrationTest(BaseWebTest, unittest.TestCase):

    channel_url = '/channels/food'
    channel_registration_url = '/channels/food/registration'

    def setUp(self):
        super(ChannelRegistrationTest, self).setUp()
        self.error = storage_exceptions.BackendError(ValueError())

    def test_registration_can_be_added_for_a_user(self):
        self.app.put(self.channel_registration_url, headers=self.headers, status=202)
        resp = self.app.get(self.channel_url, headers=self.headers, status=200)
        assert resp.json['data']['registrations'] == 1

    def test_remove_non_existent_registration(self):
        self.app.delete(self.channel_registration_url, headers=self.headers, status=202)

    def test_we_cannot_register_an_anonymous_user(self):
        self.app.put(self.channel_registration_url, status=401)

    def test_we_cannot_unregister_an_anonymous_user(self):
        self.app.delete(self.channel_registration_url, status=401)

    def test_backend_errors_are_served_as_503_on_registration(self):
        with mock.patch('kinto.core.storage.memory.Storage.update', side_effect=self.error):
            self.app.put(self.channel_registration_url,
                         headers=self.headers,
                         status=503)

    def test_backend_errors_are_served_as_503_on_channel(self):
        with mock.patch('kinto.core.storage.memory.Storage.get_all', side_effect=self.error):
            self.app.get(self.channel_url,
                         headers=self.headers,
                         status=503)


class EmptyChannelsTest(BaseWebTest, unittest.TestCase):

    channel_url = '/channels/food'
    channel_registration_url = '/channels/food/registration'

    def test_push_notifications_can_be_sent_to_channel_even_without_anyone_being_registered(self):
        self.app.post(self.channel_url, headers=self.headers, status=202)

    def test_we_cannot_access_the_channel_information_anonymously(self):
        self.app.get(self.channel_url, status=401)

    def test_we_cannot_access_the_channel_information_if_we_are_not_registered(self):
        self.app.get(self.channel_url, headers=self.headers, status=403)


class RegisteredChannelsTest(BaseWebTest, unittest.TestCase):

    channel_url = '/channels/food'
    channel_registration_url = '/channels/food/registration'
    subscription_url = '/subscriptions'

    def setUp(self):
        super(RegisteredChannelsTest, self).setUp()
        self.app.put(self.channel_registration_url, headers=self.headers, status=202)

    def test_channels_information_can_be_retrieved(self):
        resp = self.app.get(self.channel_url, headers=self.headers, status=200)
        assert 'data' in resp.json

        assert 'registrations' in resp.json['data']
        assert 'push' in resp.json['data']

        assert resp.json['data']['registrations'] == 1
        assert resp.json['data']['push'] == 0

    def test_registration_count_updated(self):
        headers = self.headers.copy()
        headers.update(testing.get_user_headers('natim'))
        self.app.put(self.channel_registration_url, headers=headers, status=202)
        resp = self.app.get(self.channel_url, headers=self.headers, status=200)
        assert resp.json['data']['registrations'] == 2

    def test_registration_can_be_remove_for_a_user(self):
        headers = self.headers.copy()
        headers.update(testing.get_user_headers('natim'))
        self.app.put(self.channel_registration_url, headers=headers, status=202)
        self.app.delete(self.channel_registration_url, headers=headers, status=202)
        resp = self.app.get(self.channel_url, headers=self.headers, status=200)
        assert resp.json['data']['registrations'] == 1

    def test_push_notifications_can_be_sent_to_channel_with_registration_but_no_subscription(self):
        self.app.post(self.channel_url, headers=self.headers, status=202)


class RegisteredAndSubscribedChannelsTest(BaseWebTest, unittest.TestCase):
    channel_url = '/channels/food'
    channel_registration_url = '/channels/food/registration'
    subscription_url = '/subscriptions'

    def setUp(self):
        super(RegisteredAndSubscribedChannelsTest, self).setUp()
        self.app.put(self.channel_registration_url, headers=self.headers, status=202)
        resp = self.app.post_json(self.subscription_url,
                                  MINIMALIST_SUBSCRIPTION,
                                  headers=self.headers)
        self.subscription = resp.json['data']
        del self.subscription['id']
        del self.subscription['last_modified']
        self.subscription['keys']['auth'] = self.subscription['keys']['auth'].encode('utf-8')
        self.subscription['keys']['p256dh'] = self.subscription['keys']['p256dh'].encode('utf-8')

        self.webpusher_error_patcher = mock.patch('webpush_channels.views.channels.WebPusher')

    def test_push_notifications_can_be_sent_with_no_payload(self):
        with self.webpusher_error_patcher as webpusher_mock:
            self.app.post(self.channel_url, headers=self.headers, status=202)
            webpusher_mock.assert_called_with(self.subscription)
            webpusher_mock.return_value.send.assert_called_with(data=None, ttl=15)

    def test_push_notification_can_take_a_payload(self):
        with self.webpusher_error_patcher as webpusher_mock:
            self.app.post_json(self.channel_url, MINIMALIST_PAYLOAD,
                               headers=self.headers, status=202)
            webpusher_mock.assert_called_with(self.subscription)
            webpusher_mock.return_value.send.assert_called_with(
                data=canonical_json(MINIMALIST_PAYLOAD['data']), ttl=15)

    def test_invalid_encryption_keys_shows_error(self):
        CHANGED_SUBSCRIPTION = deepcopy(MINIMALIST_SUBSCRIPTION)
        CHANGED_SUBSCRIPTION['data']['keys']['p256dh'] = 'yAB'

        self.app.post_json(self.subscription_url,
                           CHANGED_SUBSCRIPTION,
                           headers=self.headers, status=400)


class AllResponsesAreJSONTest(testing.FormattedErrorMixin, BaseWebTest, unittest.TestCase):

    channel_url = '/channels/food'
    invalid_channel_url = '/channels/blah'
    channel_registration_url = '/channels/food/registration'
    invalid_channel_registration_url = '/channels/blah/registration'
    subscription_url = '/subscriptions'

    def assertJSONAcceptedResponse(self, response):
        assert response.headers['Content-Type'] == 'application/json'
        assert response.json['code'] == 202
        assert response.json['message'] == 'Accepted'

    def setUp(self):
        super(AllResponsesAreJSONTest, self).setUp()
        self.app.post_json(self.subscription_url,
                           MINIMALIST_SUBSCRIPTION,
                           headers=self.headers)
        self.resp = self.app.put(self.channel_registration_url, headers=self.headers, status=202)

    def test_put_request_response_is_json(self):
        self.assertJSONAcceptedResponse(self.resp)

    def test_delete_request_response_is_json(self):
        resp = self.app.delete(self.channel_registration_url, headers=self.headers, status=202)
        self.assertJSONAcceptedResponse(resp)

    def test_get_request_response_is_json(self):
        resp = self.app.get(self.channel_url, headers=self.headers, status=200)
        assert json.loads(resp.body)
        assert resp.headers['Content-Type'] == 'application/json'

    def test_post_request_response_is_json(self):
        resp = self.app.post(self.channel_url, headers=self.headers, status=202)
        self.assertJSONAcceptedResponse(resp)

    def test_put_request_on_invalid_channel_registration_url_response_is_json(self):
        resp = self.app.put(self.channel_url, headers=self.headers, status=405)
        self.assertFormattedError(resp, 405, ERRORS.METHOD_NOT_ALLOWED, "Method Not Allowed",
                                  "Method not allowed on this endpoint.")

    def test_delete_request_on_invalid_registration_url_response_is_json(self):
        resp = self.app.delete(self.channel_url, headers=self.headers, status=405)
        self.assertFormattedError(resp, 405, ERRORS.METHOD_NOT_ALLOWED, "Method Not Allowed",
                                  "Method not allowed on this endpoint.")

    def test_delete_request_on_non_existent_registration_url_response_is_json(self):
        resp = self.app.delete(self.invalid_channel_registration_url,
                               headers=self.headers, status=202)
        self.assertJSONAcceptedResponse(resp)

    def test_get_request_on_non_existent_channel_url_response_is_json(self):
        resp = self.app.get(self.invalid_channel_url, headers=self.headers, status=403)
        self.assertFormattedError(resp, 403, ERRORS.FORBIDDEN, "Forbidden",
                                  "This user cannot access this resource.")

    def test_post_request_to_non_existent_channel_response_is_json(self):
        resp = self.app.post(self.invalid_channel_url, headers=self.headers, status=202)
        self.assertJSONAcceptedResponse(resp)

    def test_post_request_to_invalid_channel_response_is_json(self):
        resp = self.app.post(self.invalid_channel_registration_url,
                             headers=self.headers,
                             status=405)
        self.assertFormattedError(resp, 405, ERRORS.METHOD_NOT_ALLOWED, "Method Not Allowed",
                                  "Method not allowed on this endpoint.")



import unittest

from webpush_channels.views.channels import PayloadSchema

from .support import MINIMALIST_PAYLOAD


class PayloadSchemaTest(unittest.TestCase):
    def setUp(self):
        self.schema = PayloadSchema()
        self.schema = self.schema.bind()

    def test_record_validation(self):
        deserialized = self.schema.deserialize(MINIMALIST_PAYLOAD)
        self.assertEqual(deserialized, MINIMALIST_PAYLOAD)

    def test_empty_record_validation(self):
        deserialized = self.schema.deserialize({})
        self.assertEqual(deserialized, {})



from kinto.core import testing
from webpush_channels import main as webpush_main
from webpush_channels import DEFAULT_SETTINGS

P256DH = "BEVoH6cOlNPuvYR0aVJo4GVv84nbymzpXxNff7hpKYjVIFcuIEtqiLtIe4rLOXF_A2w3KWRJoCYJEjUedrXcNpc"

MINIMALIST_SUBSCRIPTION = {
    "data": {
        "endpoint": "https://updates.push.services.mozilla.com/wpush/v1/gAAAAABYZNoTAeA9vv-_zHx79",
        "keys": {
            "auth": "pnipzxpMvKBNYZAcxc-MAA",
            "p256dh": P256DH
        }
    }
}

MINIMALIST_PAYLOAD = {
    'data': {
        'url': "https://push.mozilla.com",
        'author': {
            'name': 'foobar'
        }
    }
}


class BaseWebTest(testing.BaseWebTest):

    api_prefix = "v0"
    entry_point = webpush_main

    def __init__(self, *args, **kwargs):
        super(BaseWebTest, self).__init__(*args, **kwargs)
        self.headers.update(testing.get_user_headers('mkaur'))

    def get_app_settings(self, extras=None):
        settings = DEFAULT_SETTINGS.copy()
        if extras is not None:
            settings.update(extras)
        settings = super(BaseWebTest, self).get_app_settings(extras=settings)
        return settings



import hashlib

import colander
from pyramid.httpexceptions import HTTPBadRequest
from pywebpush import WebPusher, WebPushException

from kinto.core.errors import http_error, ERRORS
from kinto.core.resource import register, UserResource, ResourceSchema
from kinto.core.storage import generators

from ..utils import canonical_json


def generate_id(subscription):
    return hashlib.sha256(canonical_json(subscription).encode('utf-8')).hexdigest()


class KeySchema(colander.MappingSchema):
    auth = colander.SchemaNode(colander.String())
    p256dh = colander.SchemaNode(colander.String())


class SubscriptionSchema(ResourceSchema):
    id = colander.SchemaNode(colander.String(), missing=colander.drop)
    endpoint = colander.SchemaNode(colander.String(), validator=colander.url)
    keys = KeySchema()

    def deserialize(self, cstruct):
        """Preprocess received data to make sure if an id is present it has
        the correct value.

        """
        record = super(SubscriptionSchema, self).deserialize(cstruct)
        given_id = record.get('id')
        generated_id = generate_id(record['endpoint'])

        if given_id and given_id != generated_id:
            raise colander.Invalid(
                self, msg="Invalid ID: '{}' found while it should be '{}'".format(
                    given_id, generated_id))

        return super(SubscriptionSchema, self).deserialize(cstruct)


class SHA256Generator(generators.Generator):

    regexp = r'^[a-zA-Z0-9][a-zA-Z0-9_-]*$'

    def __init__(self, request, config=None):
        self.request = request
        super(SHA256Generator, self).__init__(config)

    def __call__(self):
        if 'body' not in self.request.validated:
            return generate_id(None)

        body = self.request.validated['body']['data']
        return generate_id(body['endpoint'])


@register(name='subscription',
          collection_path='/subscriptions',
          record_path='/subscriptions/{{id}}')
class Subscription(UserResource):
    schema = SubscriptionSchema
    preserve_unknown = False

    def __init__(self, request, *args, **kwargs):
        super(Subscription, self).__init__(request, *args, **kwargs)
        self.model.id_generator = SHA256Generator(request)

    def process_record(self, new, old=None):
        new = super(Subscription, self).process_record(new, old)
        try:
            WebPusher(new)
        except WebPushException as e:
            raise http_error(HTTPBadRequest(),
                             errno=ERRORS.INVALID_PARAMETERS,
                             message='Invalid subscription: %s' % e)
        return new



import pkg_resources
import logging

from pyramid.config import Configurator

import kinto.core


# Module version, as defined in PEP-0396.
__version__ = pkg_resources.get_distribution(__package__).version

# The API version is derivated from the module version.
API_VERSION = 'v%s' % __version__.split('.')[0]

# Implemented HTTP API Version
HTTP_API_VERSION = '0.1'

# Main readinglist logger
logger = logging.getLogger(__name__)


DEFAULT_SETTINGS = {
    'kinto.core.paginate_by': 100,
    'flush_endpoint_enabled': False,
    'retry_after_seconds': 3,
    'cache_backend': 'kinto.core.cache.memory',
    'permission_backend': 'kinto.core.permission.memory',
    'storage_backend': 'kinto.core.storage.memory',
    'project_docs': 'https://webpush-channels-broadcasting.readthedocs.io/',
    'multiauth.authorization_policy': (
        'kinto.authorization.AuthorizationPolicy'),
    'multiauth.policies': 'basicauth',
    'multiauth.policy.basicauth.use': ('kinto.core.authentication.'
                                       'BasicAuthAuthenticationPolicy'),
    'http_api_version': HTTP_API_VERSION,
}


def main(global_config, config=None, **settings):
    if not config:
        config = Configurator(settings=settings)

    config.add_settings({'kinto.project_name': 'webpush-channels'})

    kinto.core.initialize(config,
                          version=__version__,
                          default_settings=DEFAULT_SETTINGS)

    config.scan("webpush_channels.views")
    app = config.make_wsgi_app()
    return kinto.core.install_middlewares(app, settings)



from collections import OrderedDict
from webpush_channels.utils import canonical_json


def test_canonical_json_sorts_keys():
    d1 = OrderedDict()
    d1['foobar'] = 2
    d1['dodo'] = 1
    d2 = OrderedDict()
    d2['dodo'] = 1
    d2['foobar'] = 2

    assert canonical_json(d1) == canonical_json(d2)



import os
from setuptools import setup, find_packages
import codecs

here = os.path.abspath(os.path.dirname(__file__))

def read_file(filename):
    """Open a related file and return its content."""
    with codecs.open(os.path.join(here, filename), encoding='utf-8') as f:
        content = f.read()
    return content

README = read_file("README.rst")
CHANGELOG = read_file("CHANGELOG.rst")

REQUIREMENTS = [
    'waitress',
    'kinto[postgresql]<6',
    'requests<2.13.0',
    'pywebpush>=0.7.0'
]

ENTRY_POINTS = {
    'paste.app_factory': [
        'main = webpush_channels:main',
    ]}

setup(name='webpush-channels',
      version='0.1.0',
      description='Broadcast notifications to multiple WebPush subscribers.',
      long_description=README + "\n\n" + CHANGELOG,
      license='Apache License (2.0)',
      keywords="webpush notifications, webpush channels",
      author='Mansimar Kaur',
      author_email='mansimarkaur.mks@gmail.com',
      url='https://github.com/webpush-channels/webpush-channels',
      packages=find_packages(),
      include_package_data=True,
      zip_safe=False,
      install_requires=REQUIREMENTS,
      entry_points=ENTRY_POINTS)



import json
import unittest
import uuid

from copy import deepcopy

from kinto.core.utils import decode_header
from kinto.core.testing import get_user_headers

from .support import BaseWebTest, MINIMALIST_SUBSCRIPTION


class SubscriptionsViewTest(BaseWebTest, unittest.TestCase):

    collection_url = '/subscriptions'
    _record_url = '/subscriptions/%s'

    def setUp(self):
        super(SubscriptionsViewTest, self).setUp()
        resp = self.app.post_json(self.collection_url,
                                  MINIMALIST_SUBSCRIPTION,
                                  headers=self.headers)
        self.subscription = resp.json['data']
        self.subscription_url = self._record_url % self.subscription['id']

    def test_subscriptions_can_be_accessed_by_id(self):
        self.app.get(self.subscription_url, headers=self.headers)

    def test_unknown_subscription_raises_404(self):
        new_id = '%s' % uuid.uuid4()
        other_record = self.subscription_url.replace(self.subscription['id'],
                                                     new_id)
        response = self.app.get(other_record, headers=self.headers, status=404)
        self.assertEqual(response.json['details']['id'], new_id)
        self.assertEqual(response.json['details']['resource_name'], 'subscription')

    def test_subscriptions_can_be_added(self):
        response = self.app.get(self.subscription_url, headers=self.headers)
        record = response.json['data']
        del record['id']
        del record['last_modified']
        self.assertEquals(record, MINIMALIST_SUBSCRIPTION['data'])

    def test_subscriptions_can_be_filtered_on_any_field(self):
        self.app.post_json(self.collection_url,
                           MINIMALIST_SUBSCRIPTION,
                           headers=self.headers)
        response = self.app.get(self.collection_url + '?unknown=1',
                                headers=self.headers)
        self.assertEqual(len(response.json['data']), 0)

    def test_subscriptions_can_be_sorted_on_any_field(self):
        for i in range(3):
            record = deepcopy(MINIMALIST_SUBSCRIPTION)
            record['data']['endpoint'] = 'http://endpoint/%s' % i
            record['data']['keys']['auth'] = 'Auths%s' % i
            self.app.post_json(self.collection_url,
                               record,
                               headers=self.headers)

        response = self.app.get(self.collection_url + '?_sort=-keys.auth',
                                headers=self.headers)
        names = [i['keys']['auth'] for i in response.json['data']]
        self.assertEqual(names,
                         ['pnipzxpMvKBNYZAcxc-MAA', 'Auths2', 'Auths1', 'Auths0'])

    def test_create_a_subscription_update_collection_timestamp(self):
        collection_resp = self.app.get(self.collection_url,
                                       headers=self.headers)
        old_timestamp = int(
            decode_header(json.loads(collection_resp.headers['ETag'])))
        self.app.post_json(self.collection_url,
                           MINIMALIST_SUBSCRIPTION,
                           headers=self.headers,
                           status=201)
        collection_resp = self.app.get(self.collection_url,
                                       headers=self.headers)
        new_timestamp = int(
            decode_header(json.loads(collection_resp.headers['ETag'])))
        assert old_timestamp < new_timestamp

    def test_invalid_id_of_subscription_should_not_be_accepted(self):
        record = {'data': dict(id='a-simple-id', **MINIMALIST_SUBSCRIPTION['data'])}
        self.app.post_json(self.collection_url,
                           record,
                           headers=self.headers,
                           status=400)

    def test_create_a_subscription_with_an_id_does_not_uses_it(self):
        new_id = '%s' % uuid.uuid4()
        record = {'data': dict(id=new_id, **MINIMALIST_SUBSCRIPTION['data'])}
        self.app.post_json(self.collection_url,
                           record,
                           headers=self.headers,
                           status=400)

    def test_create_a_subscription_with_an_existing_id_returns_existing(self):
        resp = self.app.post_json(self.collection_url,
                                  MINIMALIST_SUBSCRIPTION,
                                  headers=self.headers,
                                  status=201)
        existing_id = resp.json['data']['id']
        record = deepcopy(MINIMALIST_SUBSCRIPTION)
        record['data']['id'] = existing_id
        resp = self.app.post_json(self.collection_url,
                                  record,
                                  headers=self.headers,
                                  status=200)

    def test_create_a_subscription_with_existing_from_someone_else_gives_201(self):
        resp = self.app.post_json(self.collection_url,
                                  MINIMALIST_SUBSCRIPTION,
                                  headers=self.headers,
                                  status=201)
        existing_id = resp.json['data']['id']
        record = deepcopy(MINIMALIST_SUBSCRIPTION)
        record['data']['id'] = existing_id
        resp = self.app.post_json(self.collection_url,
                                  record,
                                  headers=get_user_headers('tartanpion'),
                                  status=201)

    def test_update_a_subscription_update_collection_timestamp(self):
        collection_resp = self.app.get(self.collection_url,
                                       headers=self.headers)
        old_timestamp = int(
            decode_header(json.loads(collection_resp.headers['ETag'])))
        self.app.put_json(self.subscription_url,
                          MINIMALIST_SUBSCRIPTION,
                          headers=self.headers,
                          status=200)
        collection_resp = self.app.get(self.collection_url,
                                       headers=self.headers)
        new_timestamp = int(
            decode_header(json.loads(collection_resp.headers['ETag'])))
        assert old_timestamp < new_timestamp

    def test_delete_a_subscription_update_collection_timestamp(self):
        collection_resp = self.app.get(self.collection_url,
                                       headers=self.headers)
        old_timestamp = int(
            decode_header(json.loads(collection_resp.headers['ETag'])))
        self.app.delete(self.subscription_url,
                        headers=self.headers,
                        status=200)
        collection_resp = self.app.get(self.collection_url,
                                       headers=self.headers)
        new_timestamp = int(
            decode_header(json.loads(collection_resp.headers['ETag'])))
        assert old_timestamp < new_timestamp

    def test_subscriptions_should_reject_unaccepted_request_content_type(self):
        headers = self.headers.copy()
        headers['Content-Type'] = 'text/plain'
        self.app.put(self.subscription_url,
                     MINIMALIST_SUBSCRIPTION,
                     headers=headers,
                     status=415)

    def test_subscriptions_should_reject_unaccepted_client_accept(self):
        headers = self.headers.copy()
        headers['Accept'] = 'text/plain'
        self.app.get(self.subscription_url,
                     MINIMALIST_SUBSCRIPTION,
                     headers=headers,
                     status=406)

    def test_subscriptions_should_accept_client_accept(self):
        headers = self.headers.copy()
        headers['Accept'] = '*/*'
        self.app.get(self.subscription_url,
                     MINIMALIST_SUBSCRIPTION,
                     headers=headers,
                     status=200)

    def test_subscriptions_can_be_created_after_deletion(self):
        self.app.delete(self.subscription_url,
                        headers=self.headers,
                        status=200)
        headers = self.headers.copy()
        headers['If-None-Match'] = '*'
        self.app.put_json(self.subscription_url, MINIMALIST_SUBSCRIPTION,
                          headers=headers, status=201)

    def test_multiple_subscriptions_merged(self):
        resp = self.app.post_json(self.collection_url,
                                  MINIMALIST_SUBSCRIPTION,
                                  headers=self.headers,
                                  status=201)
        subscription = resp.json['data']

        # XXX: Right now Kinto.core doesn't return the correct status code
        resp = self.app.post_json(self.collection_url,
                                  MINIMALIST_SUBSCRIPTION,
                                  headers=self.headers,
                                  status=201)
        new_subscription = resp.json['data']
        del new_subscription['last_modified']
        del subscription['last_modified']
        assert new_subscription == subscription


class AllResponsesAreJSONTest(BaseWebTest, unittest.TestCase):

    subscription_url = '/subscriptions'

    def setUp(self):
        self.resp = self.app.post_json(self.subscription_url,
                                       MINIMALIST_SUBSCRIPTION,
                                       headers=self.headers,
                                       status=201)

    def test_post_request_response_is_json(self):
        assert json.loads(self.resp.body)
        assert self.resp.headers['Content-Type'] == 'application/json'

    def test_get_request_response_is_json(self):
        resp = self.app.get(self.subscription_url,
                            headers=self.headers,
                            status=200)
        assert json.loads(resp.body)
        assert resp.headers['Content-Type'] == 'application/json'

    def test_delete_request_response_is_json(self):
        resp = self.app.delete(self.subscription_url,
                               headers=self.headers,
                               status=200)
        assert json.loads(resp.body)
        assert resp.headers['Content-Type'] == 'application/json'

    def test_delete_specific_subscription_request_response_is_json(self):
        resp = self.app.delete(self.subscription_url+'/'+self.resp.json['data']['id'],
                               headers=self.headers,
                               status=200)
        assert json.loads(resp.body)
        assert resp.headers['Content-Type'] == 'application/json'

    def test_get_request_for_unknown_subscription_url_response_is_json(self):
        resp = self.app.get('/foo',
                            headers=self.headers,
                            status=404)
        assert json.loads(resp.body)
        assert resp.headers['Content-Type'] == 'application/json'

    def test_post_invalid_subscription_response_is_json(self):
        INVALID_SUBSCRIPTION = deepcopy(MINIMALIST_SUBSCRIPTION)
        INVALID_SUBSCRIPTION['keys'] = ''
        resp = self.app.post_json(self.subscription_url,
                                  INVALID_SUBSCRIPTION,
                                  headers=self.headers,
                                  status=400)
        assert json.loads(resp.body)
        assert resp.headers['Content-Type'] == 'application/json'

    def test_delete_request_for_unknown_subscription_url_response_is_json(self):
        resp = self.app.delete('/foo',
                               headers=self.headers,
                               status=404)
        assert json.loads(resp.body)
        assert resp.headers['Content-Type'] == 'application/json'

    def test_delete_nonexistent_subscription_request_response_is_json(self):
        resp = self.app.delete(self.subscription_url+'/'+'blah',
                               headers=self.headers,
                               status=404)
        assert json.loads(resp.body)
        assert resp.headers['Content-Type'] == 'application/json'



import colander
from cornice.validators import colander_body_validator
from pyramid import httpexceptions

from pywebpush import WebPusher

from kinto.core import Service
from kinto.core.resource.viewset import StrictSchema, SimpleSchema
from kinto.core.storage.exceptions import RecordNotFoundError
from kinto.core.authorization import PRIVATE

from ..utils import canonical_json


REGISTRATION_COLLECTION_ID = 'channel_registration'
SUBSCRIPTION_COLLECTION_ID = 'subscription'

channel = Service(name='channel',
                  description='Handle channel information',
                  path='/channels/{channel_id}')


channel_registration = Service(name='channel_registration',
                               description='Handle user registration for a channel',
                               path='/channels/{channel_id}/registration')


class PayloadSchema(StrictSchema):
    data = SimpleSchema(missing=colander.drop)


# Channel views
@channel.post(permission=PRIVATE, schema=PayloadSchema(), validators=(colander_body_validator,))
def send_push_notifications(request):
    channel_id = request.matchdict['channel_id']
    parent_id = '/channels/{}'.format(channel_id)

    registrations, count = request.registry.storage.get_all(
        collection_id=REGISTRATION_COLLECTION_ID,
        parent_id=parent_id)

    subscriptions = []

    for registration in registrations:
        user_subscriptions, count = request.registry.storage.get_all(
            collection_id=SUBSCRIPTION_COLLECTION_ID,
            parent_id=registration['id'])
        subscriptions += user_subscriptions

    for subscription in subscriptions:
        del subscription['id']
        del subscription['last_modified']
        data = request.validated.get('data')
        if data:
            data = canonical_json(data)
        push_initialize = WebPusher(subscription)
        push_initialize.send(data=data, ttl=15)

    request.response.status = 202
    return {"code": 202, "message": "Accepted"}


@channel.get(permission=PRIVATE)
def retrieve_channel_information(request):
    channel_id = request.matchdict['channel_id']
    parent_id = '/channels/{}'.format(channel_id)

    registrations, count = request.registry.storage.get_all(
        collection_id=REGISTRATION_COLLECTION_ID,
        parent_id=parent_id)

    user_registered = [r for r in registrations if r['id'] == request.prefixed_userid]

    if not user_registered:
        raise httpexceptions.HTTPForbidden()

    return {"data": {
        "registrations": count,
        "push": 0
    }}


# Channel Registration views
@channel_registration.put(permission=PRIVATE)
def add_user_registration(request):
    channel_id = request.matchdict['channel_id']
    parent_id = '/channels/{}'.format(channel_id)

    request.registry.storage.update(
        collection_id=REGISTRATION_COLLECTION_ID,
        parent_id=parent_id,
        object_id=request.prefixed_userid,
        record={})

    request.response.status = 202
    return {"code": 202, "message": "Accepted"}


@channel_registration.delete(permission=PRIVATE)
def remove_user_registration(request):
    channel_id = request.matchdict['channel_id']
    parent_id = '/channels/{}'.format(channel_id)

    try:
        request.registry.storage.delete(
            collection_id=REGISTRATION_COLLECTION_ID,
            parent_id=parent_id,
            object_id=request.prefixed_userid,
            with_deleted=True)
    except RecordNotFoundError:
        # If the record has already been removed that's fine.
        pass
    request.response.status = 202
    return {"code": 202, "message": "Accepted"}
